using System.Diagnostics;
using System.Data;
using System.Collections;
using Microsoft.VisualBasic;
using System.Collections.Generic;
using System;

namespace Optim_Marquardt1
{
	public delegate double MyFxDelegate(int nNumVars, ref double[] fX, ref double[] fParam);
	public delegate string SayFxDelegate();
	
	public class CMarquardt1
	{
		
		MyFxDelegate m_MyFx;
		
		protected double MyFxEx(int nNumVars, ref double[] fX, ref double[] fParam, ref double[] fDeltaX, double fLambda)
		{
			int i;
			double[] fXX = new double[nNumVars];
			
			for (i = 0; i < nNumVars; i++)
			{
				fXX[i] = fX[i] + fLambda * fDeltaX[i];
			}
			
			return m_MyFx(nNumVars, ref fXX, ref fParam);
		}
		
		protected double FirstDeriv(int nNumVars, ref double[] fX, ref double[] fParam, int nIdxI)
		{
			double fXt, h, Fp, Fm;
			
			fXt = fX[nIdxI];
			h = 0.01 *(1 + Math.Abs(fXt));
			fX[nIdxI] = fXt + h;
			Fp = m_MyFx(nNumVars, ref fX, ref fParam);
			fX[nIdxI] = fXt - h;
			Fm = m_MyFx(nNumVars, ref fX, ref fParam);
			fX[nIdxI] = fXt;
			return (Fp - Fm) / 2 / h;
		}
		
		protected double SecondDeriv(int nNumVars, ref double[] fX, ref double[] fParam, int nIdxI, int nIdxJ)
		{
			double fXt, fYt, fHX, fHY, F0, Fp, Fm;
			double Fpp, Fmm, Fpm, Fmp, fResult;
			
			// calculate second derivative?
			if (nIdxI == nIdxJ)
			{
				F0 = m_MyFx(nNumVars, ref fX, ref fParam);
				fXt = fX[nIdxI];
				fHX = 0.01 *(1 + Math.Abs(fXt));
				fX[nIdxI] = fXt + fHX;
				Fp = m_MyFx(nNumVars, ref fX, ref fParam);
				fX[nIdxI] = fXt - fHX;
				Fm = m_MyFx(nNumVars, ref fX, ref fParam);
				fX[nIdxI] = fXt;
				fResult = (Fp - 2 * F0 + Fm) /Math.Pow(fHX, 2);
			}
			else
			{
				fXt = fX[nIdxI];
				fYt = fX[nIdxJ];
				fHX = 0.01 *(1 + Math.Abs(fXt));
				fHY = 0.01 *(1 + Math.Abs(fYt));
				// calculate Fpp
				fX[nIdxI] = fXt + fHX;
				fX[nIdxJ] = fYt + fHY;
				Fpp = m_MyFx(nNumVars, ref fX, ref fParam);
				// calculate Fmm
				fX[nIdxI] = fXt - fHX;
				fX[nIdxJ] = fYt - fHY;
				Fmm = m_MyFx(nNumVars, ref fX, ref fParam);
				// calculate Fpm
				fX[nIdxI] = fXt + fHX;
				fX[nIdxJ] = fYt - fHY;
				Fpm = m_MyFx(nNumVars, ref fX, ref fParam);
				// calculate Fmp
				fX[nIdxI] = fXt - fHX;
				fX[nIdxJ] = fYt + fHY;
				Fmp = m_MyFx(nNumVars, ref fX, ref fParam);
				fX[nIdxI] = fXt;
				fX[nIdxJ] = fYt;
				fResult = (Fpp - Fmp - Fpm + Fmm) /(4 * fHX * fHY);
			}
			
			return fResult;
		}
		
		
		protected void GetFirstDerives(int nNumVars, ref double[] fX, ref double[] fParam, ref double[] fFirstDerivX)
		{
			
			int i;
			
			for (i = 0; i < nNumVars; i++)
			{
				fFirstDerivX[i] = FirstDeriv(nNumVars, ref fX, ref fParam, i);
			}
		}
		
		protected void GetSecondDerives(int nNumVars, ref double[] fX, ref double[] fParam, ref double[,] fSecondDerivX)
		{
			
			int i, j;
			
			for (i = 0; i < nNumVars; i++)
			{
				for (j = 0; j < nNumVars; j++)
				{
					fSecondDerivX[i, j] = SecondDeriv(nNumVars, ref fX, ref fParam, i, j);
				}
			}
		}
		
		protected bool LinSearch_DirectSearch(int nNumVars, ref double[] fX, ref double[] fParam, ref double fLambda, ref double[] fDeltaX, double fInitStep, double fMinStep)
		{
			double F1, F2;

            F1 = MyFxEx(nNumVars, ref fX, ref fParam, ref fDeltaX, fLambda);
			
			do
			{
                F2 = MyFxEx(nNumVars, ref fX, ref fParam, ref fDeltaX, fLambda + fInitStep);
				if (F2 < F1)
				{
					F1 = F2;
					fLambda += fInitStep;
				}
				else
				{
                    F2 = MyFxEx(nNumVars, ref fX, ref fParam, ref fDeltaX, fLambda - fInitStep);
					if (F2 < F1)
					{
						F1 = F2;
						fLambda -= fInitStep;
					}
					else
					{
						// reduce search step size
						fInitStep /= 10;
					}
				}
			} while (!(fInitStep < fMinStep));
			
			return true;
			
		}
		
		public double CalcOptim(int nNumVars, ref double[] fX, ref double[] fParam, ref double[] fToler, double fEpsFx, int nMaxIter, ref double fAlpha, double C1, double C2, ref int nIter, ref string sErrorMsg, MyFxDelegate MyFx)
		{
			
			const double ALPHA_MAX = 1.0E+100;
			
			int i;
			double F1;
			double F2 = 0;
			double fNorm;
			int nInnerLoopIter;
			bool bStop;
			double[] fX2 = new double[nNumVars];
			double[] g = new double[nNumVars];
			double[] G2 = new double[nNumVars];
			double[] DeltaX = new double[nNumVars];
			double[,] j = new double[nNumVars, nNumVars];
			double[,] J2 = new double[nNumVars, nNumVars];
			double[,] fIdenMat = new double[nNumVars, nNumVars];
			double[,] fIdenMat2 = new double[nNumVars, nNumVars];
			int[] nIndex = new int[nNumVars];
			
			try
			{
				m_MyFx = MyFx;
				
				// validate iterations parameters
				if (C1 <= 0 || C1 >= 1)
				{
					C1 = 0.1;
				}
				if (C2 <= 1)
				{
					C2 = 10;
				}
				
				F2 = MyFx(nNumVars, ref fX, ref fParam);
				
				nIter = 1;
				do
				{
					nIter++;
					if (nIter > nMaxIter)
					{
						sErrorMsg = "Reached maximum iterations limit";
						break;
					}
					
					// copy last Fx value
					F1 = F2;

                    GetFirstDerives(nNumVars, ref fX, ref fParam, ref g);
                    MatrixLibVb.DuplicateVector(ref g, ref G2);
					
					// test if gradient is shallow enough
					fNorm = 0;
					for (i = 0; i < nNumVars; i++)
					{
						fNorm += Math.Pow(g[i], 2);
					}
					fNorm = Math.Sqrt(fNorm);
					if (fNorm < fEpsFx)
					{
						break;
					}
					
					MatrixLibVb.DuplicateVector(ref fX, ref fX2); // copy fX
					// get matrix j
					GetSecondDerives(nNumVars, ref fX, ref fParam, ref j);
					// create identitty matrix
					MatrixLibVb.MatIdentity(ref fIdenMat);
					// initialize inner-loop counter
					nInnerLoopIter = 0;
					
					do
					{
						nInnerLoopIter++;
						if (nInnerLoopIter > nMaxIter)
						{
							return 0;
						}
						
						MatrixLibVb.DuplicateVector(ref G2, ref g);
                        MatrixLibVb.MatMultVal(ref fIdenMat, ref fAlpha, ref fIdenMat2);
                        MatrixLibVb.MatAddMat(ref j, ref fIdenMat2, ref J2);
						MatrixLibVb.MV_LUDecomp(ref J2, ref nIndex, nNumVars);
						MatrixLibVb.MV_LUBackSubst(ref J2, ref nIndex, nNumVars, ref g);
						for (i = 0; i < nNumVars; i++)
						{
							DeltaX[i] = g[i];
							fX[i] -= DeltaX[i];
						}
						
						F2 = MyFx(nNumVars, ref fX, ref fParam);
						
						if (F2 < F1)
						{
							fAlpha = C1 * fAlpha;
							bStop = true;
						}
						else
						{
							fAlpha = C2 * fAlpha;
							MatrixLibVb.DuplicateVector(ref fX2, ref fX); // restore array fX
							bStop = false;
							if (fAlpha >= ALPHA_MAX)
							{
								return F2;
							}
						}
					} while (!bStop);
					
				} while (true);
			}
			catch (Exception ex)
			{
				sErrorMsg = "Calculation error: " + ex.Message;
			}
            return F2;

		}
		
	}
	
}
